#ifndef TIMESCALEDB_DIMENSION_SLICE_H
#define TIMESCALEDB_DIMENSION_SLICE_H

#include <postgres.h>
#include <nodes/pg_list.h>

#include "catalog.h"
#include "dimension.h"
#include "chunk_constraint.h"

typedef struct DimensionSlice
{
	FormData_dimension_slice fd;
	void		(*storage_free) (void *);
	void	   *storage;
} DimensionSlice;

typedef struct DimensionVec DimensionVec;
typedef struct Hypercube Hypercube;

extern DimensionVec *dimension_slice_scan_limit(int32 dimension_id, int64 coordinate, int limit);
extern DimensionVec *dimension_slice_collision_scan_limit(int32 dimension_id, int64 range_start, int64 range_end, int limit);
extern Hypercube *dimension_slice_point_scan(Hyperspace *space, int64 point[]);
extern DimensionSlice *dimension_slice_scan_for_existing(DimensionSlice *slice);
extern DimensionSlice *dimension_slice_scan_by_id(int32 dimension_slice_id);
extern DimensionSlice *dimension_slice_create(int dimension_id, int64 range_start, int64 range_end);
extern DimensionSlice *dimension_slice_copy(const DimensionSlice *original);
extern bool dimension_slices_collide(DimensionSlice *slice1, DimensionSlice *slice2);
extern bool dimension_slices_equal(DimensionSlice *slice1, DimensionSlice *slice2);
extern bool dimension_slice_cut(DimensionSlice *to_cut, DimensionSlice *other, int64 coord);
extern void dimension_slice_free(DimensionSlice *slice);
extern void dimension_slice_insert_multi(DimensionSlice **slice, Size num_slices);

#define dimension_slice_insert(slice) \
	dimension_slice_insert_multi(&(slice), 1)

#define dimension_slice_scan(dimension_id, coordinate)	\
	dimension_slice_scan_limit(dimension_id, coordinate, 0)

#define dimension_slice_collision_scan(dimension_id, range_start, range_end)		\
	dimension_slice_collision_scan_limit(dimension_id, range_start, range_end, 0)


#endif   /* TIMESCALEDB_DIMENSION_SLICE_H */
